from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from kraken.util.requirements import RequirementSpec


@dataclasses.dataclass
class Lockfile:
    requirements: RequirementSpec
    pinned: dict[str, str]

    @staticmethod
    def from_path(path: Path) -> Lockfile:
        import tomli

        with path.open("rb") as fp:
            return Lockfile.from_json(tomli.load(fp))

    def write_to(self, path: Path) -> None:
        import tomli_w

        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("wb") as fp:
            tomli_w.dump(self.to_json(), fp)

    @staticmethod
    def from_json(data: dict[str, Any]) -> Lockfile:
        from kraken.util.requirements import RequirementSpec

        return Lockfile(
            requirements=RequirementSpec.from_json(data["requirements"]),
            pinned=data["pinned"],
        )

    def to_json(self) -> dict[str, Any]:
        return {
            "requirements": self.requirements.to_json(),
            "pinned": self.pinned,
        }

    def to_pinned_requirement_spec(self) -> RequirementSpec:
        """Converts the pinned versions in the lock file to a :class:`RequirementSpec` with the pinned requirements."""

        from kraken.util.requirements import LocalRequirement, RequirementSpec

        requirements = RequirementSpec(
            requirements=(),
            index_url=self.requirements.index_url,
            extra_index_urls=self.requirements.extra_index_urls[:],
            pythonpath=self.requirements.pythonpath[:],
            interpreter_constraint=self.requirements.interpreter_constraint,
        )

        # Make sure that local requirements keep being installed from the local source.
        local_requirements = {
            dep.name: dep for dep in self.requirements.requirements if isinstance(dep, LocalRequirement)
        }
        requirements = requirements.with_requirements(local_requirements.values())

        # Add all non-local requirements with exact version numbers.
        requirements = requirements.with_requirements(
            f"{key}=={value}" for key, value in sorted(self.pinned.items()) if key not in local_requirements
        )

        return requirements



from __future__ import annotations

import abc
import contextlib
import dataclasses
import datetime
import enum
import hashlib
import json
import logging
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, NoReturn, Sequence

from kraken.util.helpers import NotSet, not_none

if TYPE_CHECKING:
    from kraken.util.requirements import RequirementSpec

    from kraken.wrapper.lockfile import Lockfile

logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class Distribution:
    name: str
    version: str
    requirements: list[str]
    extras: set[str]


class BuildEnv(abc.ABC):
    """Interface for the build environment."""

    @abc.abstractmethod
    def get_type(self) -> BuildEnvType:
        """Return the type of build environment that this is."""

    @abc.abstractmethod
    def get_path(self) -> Path:
        """Return the path to the build environment."""

    @abc.abstractmethod
    def get_installed_distributions(self) -> list[Distribution]:
        """Return the distributions that are currently installed in the environment."""

    @abc.abstractmethod
    def build(self, requirements: RequirementSpec) -> None:
        """Build the environment from the given requirement spec."""

    @abc.abstractmethod
    def dispatch_to_kraken_cli(self, argv: list[str]) -> NoReturn:
        """Dispatch the kraken cli command in *argv* to the build environment.

        :param argv: The arguments to pass to the kraken cli (without the "kraken" command name itself)."""


class BuildEnvType(enum.Enum):
    PEX_ZIPAPP = enum.auto()
    PEX_PACKED = enum.auto()
    PEX_LOOSE = enum.auto()
    VENV = enum.auto()


@dataclasses.dataclass(frozen=True)
class BuildEnvMetadata:
    created_at: datetime.datetime
    environment_type: BuildEnvType
    requirements_hash: str
    hash_algorithm: str

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> BuildEnvMetadata:
        from kraken.util.json import json2dt

        return cls(
            created_at=json2dt(data["created_at"]),
            environment_type=BuildEnvType[data["environment_type"]],
            requirements_hash=data["requirements_hash"],
            hash_algorithm=data["hash_algorithm"],
        )

    def to_json(self) -> dict[str, Any]:
        from kraken.util.json import dt2json

        return {
            "created_at": dt2json(self.created_at),
            "environment_type": self.environment_type.name,
            "requirements_hash": self.requirements_hash,
            "hash_algorithm": self.hash_algorithm,
        }


@dataclasses.dataclass
class BuildEnvMetadataStore:
    path: Path

    def __post_init__(self) -> None:
        self._metadata: BuildEnvMetadata | None | NotSet = NotSet.Value

    def get(self) -> BuildEnvMetadata | None:
        if self._metadata is NotSet.Value:
            if self.path.is_file():
                self._metadata = BuildEnvMetadata.from_json(json.loads(self.path.read_text()))
            else:
                self._metadata = None
        return self._metadata

    def set(self, metadata: BuildEnvMetadata) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.write_text(json.dumps(metadata.to_json()))
        self._metadata = metadata


class PexBuildEnv(BuildEnv):
    STYLES = (BuildEnvType.PEX_ZIPAPP, BuildEnvType.PEX_PACKED, BuildEnvType.PEX_LOOSE)

    def __init__(self, style: BuildEnvType, path: Path) -> None:
        assert style in self.STYLES
        self._style = style
        self._path = path

    @contextlib.contextmanager
    def activate(self) -> Iterator[None]:
        import copy

        from pex.pex import PEX
        from pex.pex_bootstrapper import bootstrap_pex_env

        assert self._path.exists(), f'expected PEX file at "{self._path}"'
        pex = PEX(self._path)

        state = {}
        for key in ["displayhook", "excepthook", "modules", "path", "path_importer_cache"]:
            state[key] = copy.copy(getattr(sys, key))

        try:
            bootstrap_pex_env(str(pex.path()))
            pex.activate()
            yield
        finally:
            for key, value in state.items():
                setattr(sys, key, value)

    # BuildEnv

    def get_path(self) -> Path:
        return self._path

    def get_type(self) -> BuildEnvType:
        return self._style

    def get_installed_distributions(self) -> list[Distribution]:
        return _get_installed_distributions([sys.executable, str(self._path)])

    def build(self, requirements: RequirementSpec) -> None:
        import pprint

        from kraken.util.text import lazy_str

        from kraken.wrapper.pex import PEXBuildConfig, PEXLayout

        config = PEXBuildConfig(
            interpreter_constraints=(
                [requirements.interpreter_constraint] if requirements.interpreter_constraint else []
            ),
            script="kraken",
            requirements=requirements.to_args(Path.cwd(), with_options=False),
            index_url=requirements.index_url,
            extra_index_urls=list(requirements.extra_index_urls),
        )

        layout = {
            BuildEnvType.PEX_ZIPAPP: PEXLayout.ZIPAPP,
            BuildEnvType.PEX_PACKED: PEXLayout.PACKED,
            BuildEnvType.PEX_LOOSE: PEXLayout.LOOSE,
        }[self._style]

        logger.debug("PEX build configuration is %s", lazy_str(lambda: pprint.pformat(config)))

        logger.info('begin PEX resolve for build environment "%s"', self._path)
        installed = config.resolve()

        logger.info('building PEX for build environment "%s"', self._path)
        builder = config.builder(installed)
        builder.build(str(self._path), layout=layout)

    def dispatch_to_kraken_cli(self, argv: list[str]) -> NoReturn:
        with self.activate():
            from kraken.cli.main import main

            main("krakenw", argv)
        assert False


class VenvBuildEnv(BuildEnv):
    def __init__(self, path: Path) -> None:
        from kraken._vendor.nr.python.environment.virtualenv import VirtualEnvInfo

        self._path = path
        self._venv = VirtualEnvInfo(self._path)

    # BuildEnv

    def get_path(self) -> Path:
        return self._path

    def get_type(self) -> BuildEnvType:
        return BuildEnvType.VENV

    def get_installed_distributions(self) -> list[Distribution]:
        python = self._venv.get_bin("python")
        return _get_installed_distributions([str(python), "-m", "kraken.cli.main"])

    def build(self, requirements: RequirementSpec) -> None:
        command = [sys.executable, "-m", "venv", str(self._path)]
        subprocess.check_call(command)

        python_bin = str(self._venv.get_bin("python"))
        command = [python_bin, "-m", "pip", "install", *requirements.to_args()]
        subprocess.check_call(command)

        # Make sure the pythonpath from the requirements is encoded into the enviroment.
        if requirements.pythonpath:
            command = [python_bin, "-c", "from sysconfig import get_path; print(get_path('purelib'))"]
            site_packages = Path(subprocess.check_output(command).decode().strip())
            pth_file = site_packages / "krakenw.pth"
            pth_file.write_text("\n".join(str(Path(path).absolute()) for path in requirements.pythonpath))

    def dispatch_to_kraken_cli(self, argv: list[str]) -> NoReturn:
        python = self._venv.get_bin("python")
        command = [str(python), "-m", "kraken.cli.main", *argv]
        sys.exit(subprocess.call(command))


class BuildEnvManager:
    def __init__(
        self,
        path: Path,
        default_type: BuildEnvType = BuildEnvType.PEX_ZIPAPP,
        default_hash_algorithm: str = "sha256",
    ) -> None:
        from kraken.util.path import with_name

        assert (
            default_hash_algorithm in hashlib.algorithms_available
        ), f"hash algoritm {default_hash_algorithm!r} is not available"

        self._path = path
        self._metadata_store = BuildEnvMetadataStore(with_name(path, path.name + ".meta"))
        self._default_type = default_type
        self._default_hash_algorithm = default_hash_algorithm

    def exists(self) -> bool:
        if self._metadata_store.get() is None:
            return False  # If we don't have metadata, we assume the environment does not exist.
        return self.get_environment().get_path().exists()

    def remove(self) -> None:
        from kraken.util.fs import safe_rmpath

        safe_rmpath(self._metadata_store.path)
        safe_rmpath(self.get_environment().get_path())

    def install(self, requirements: RequirementSpec, env_type: BuildEnvType | None = None) -> None:
        if env_type is None:
            metadata = self._metadata_store.get()
            env_type = metadata.environment_type if metadata else self._default_type

        env = _get_environment_for_type(env_type, self._path)
        env.build(requirements)
        hash_algorithm = self.get_hash_algorithm()
        metadata = BuildEnvMetadata(
            datetime.datetime.utcnow(),
            env.get_type(),
            requirements.to_hash(hash_algorithm),
            hash_algorithm,
        )
        self._metadata_store.set(metadata)

    def get_metadata_file(self) -> Path:
        return self._metadata_store.path

    def get_metadata(self) -> BuildEnvMetadata:
        return not_none(self._metadata_store.get(), "metadata does not exist")

    def get_hash_algorithm(self) -> str:
        metadata = self._metadata_store.get()
        return metadata.hash_algorithm if metadata else self._default_hash_algorithm

    def get_environment(self) -> BuildEnv:
        metadata = self._metadata_store.get()
        environment_type = self._default_type if metadata is None else metadata.environment_type
        return _get_environment_for_type(environment_type, self._path)

    def set_locked(self, lockfile: Lockfile) -> None:
        metadata = self._metadata_store.get()
        assert metadata is not None
        metadata = BuildEnvMetadata(
            metadata.created_at,
            metadata.environment_type,
            lockfile.to_pinned_requirement_spec().to_hash(metadata.hash_algorithm),
            metadata.hash_algorithm,
        )
        self._metadata_store.set(metadata)


def _get_environment_for_type(environment_type: BuildEnvType, base_path: Path) -> BuildEnv:
    from kraken.util.path import with_name

    if environment_type in PexBuildEnv.STYLES:
        return PexBuildEnv(environment_type, with_name(base_path, base_path.name + ".pex"))
    elif environment_type == BuildEnvType.VENV:
        return VenvBuildEnv(base_path)
    else:
        raise RuntimeError(f"unsupported environment type: {environment_type!r}")


def _get_installed_distributions(kraken_command_prefix: Sequence[str]) -> list[Distribution]:
    command = [*kraken_command_prefix, "query", "env"]
    output = subprocess.check_output(command).decode()
    return [Distribution(x["name"], x["version"], x["requirements"], x["extras"]) for x in json.loads(output)]



""" A wrapper around the PEX builder interface. """
from __future__ import annotations

import dataclasses
from pathlib import Path

from pex.pex import PEX
from pex.pex_builder import CopyMode, Layout as PEXLayout, PEXBuilder
from pex.platforms import Platform
from pex.resolve.resolver_configuration import PYPI, PasswordEntry, ResolverVersion
from pex.resolve.resolvers import Installed
from pex.resolver import resolve
from pex.targets import Targets
from pex.variables import Variables as PEXVariables

__all__ = [
    "CopyMode",
    "PasswordEntry",
    "PEX",
    "PEXBuildConfig",
    "PEXLayout",
    "PEXVariables",
    "ResolverVersion",
]


@dataclasses.dataclass
class PEXBuildConfig:

    # PEXBuilder
    preamble: str | None = None
    copy_mode: CopyMode = CopyMode.SYMLINK
    interpreter_constraints: list[str] = dataclasses.field(default_factory=list)
    entry_point: str | None = None
    script: str | None = None
    executable: Path | None = None
    shebang: str | None = None

    # PexInfo
    venv_site_packages_copies: bool = True
    emit_warnings: bool = True
    strip_pex_env: bool = False
    use_venv: bool = False

    # Resolver
    requirements: list[str] = dataclasses.field(default_factory=list)
    index_url: str | None = None
    extra_index_urls: list[str] = dataclasses.field(default_factory=list)
    password_entries: list[PasswordEntry] = dataclasses.field(default_factory=list)
    pip_resolver_version: ResolverVersion = ResolverVersion.PIP_2020
    platforms: list[str] | None = None
    preserve_log: bool = False

    def resolve(self) -> Installed:
        if self.platforms is None:
            targets = Targets()
        else:
            targets = Targets(platforms=tuple(Platform.create(x) for x in self.platforms))
        if self.index_url is None:
            indexes = [PYPI]
        else:
            indexes = [self.index_url]
        indexes.extend(self.extra_index_urls)
        return resolve(
            targets=targets,
            requirements=self.requirements,
            indexes=indexes,
            resolver_version=self.pip_resolver_version,
            password_entries=self.password_entries,
            preserve_log=self.preserve_log,
        )

    def builder(self, installed: Installed | None) -> PEXBuilder:
        if installed is None:
            installed = self.resolve()

        pex_builder = PEXBuilder(
            preamble=self.preamble,
            copy_mode=self.copy_mode,
        )

        pex_builder.info.venv_site_packages_copies = self.venv_site_packages_copies
        pex_builder.info.emit_warnings = self.emit_warnings
        pex_builder.info.strip_pex_env = self.strip_pex_env
        pex_builder.info.venv = self.use_venv

        for ic in self.interpreter_constraints:
            pex_builder.add_interpreter_constraint(ic)
        for installed_dist in installed.installed_distributions:
            pex_builder.add_distribution(installed_dist.distribution, fingerprint=installed_dist.fingerprint)
            for direct_req in installed_dist.direct_requirements:
                pex_builder.add_requirement(direct_req)

        if self.entry_point:
            pex_builder.set_entry_point(self.entry_point)
        elif self.script:
            pex_builder.set_script(self.script)
        elif self.executable:
            pex_builder.set_executable(str(self.executable), "__pex_executable__.py")

        if self.shebang:
            pex_builder.set_shebang(self.shebang)

        return pex_builder



from __future__ import annotations

import dataclasses
import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import argparse

    from kraken.wrapper.buildenv import BuildEnvType


@dataclasses.dataclass(frozen=True)
class EnvOptions:
    status: bool
    upgrade: bool
    reinstall: bool
    uninstall: bool
    use: BuildEnvType | None

    @staticmethod
    def add_to_parser(parser: argparse.ArgumentParser) -> None:
        from kraken.wrapper.buildenv import BuildEnvType

        parser.add_argument(
            "--status",
            action="store_true",
            help="print the status of the build environment and exit",
        )
        parser.add_argument(
            "--upgrade",
            action="store_true",
            help="reinstall the build environment from the original requirements",
        )
        parser.add_argument(
            "--reinstall",
            action="store_true",
            help="reinstall the build environment from the lock file",
        )
        parser.add_argument(
            "--uninstall",
            action="store_true",
            help="uninstall the build environment",
        )
        parser.add_argument(
            "--use",
            choices=[v.name for v in BuildEnvType],
            default=os.getenv("KRAKENW_USE"),
            help="use the specified environment type. If the environment type changes it will trigger a reinstall.\n"
            "Defaults to the value of the KRAKENW_USE environment variable. If that variable is unset, and\nif a build "
            "environment already exists, that environment's type will be used. The default\nenvironment type that is "
            "used for new environments is PEX_ZIPAPP.",
        )

    @classmethod
    def collect(cls, args: argparse.Namespace) -> EnvOptions:
        from kraken.wrapper.buildenv import BuildEnvType

        return cls(
            status=args.status,
            upgrade=args.upgrade,
            reinstall=args.reinstall,
            uninstall=args.uninstall,
            use=BuildEnvType[args.use] if args.use else None,
        )

    def any(self) -> bool:
        return bool(self.status or self.upgrade or self.reinstall or self.uninstall or self.use)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['wrapper']

package_data = \
{'': ['*']}

install_requires = \
['kraken-core>=0.8.2,<0.9.0',
 'pex>=2.1.103,<3.0.0',
 'tomli>=2.0.1,<3.0.0',
 'tomli_w>=1.0.0,<2.0.0']

entry_points = \
{'console_scripts': ['krakenw = kraken.wrapper.main:main']}

setup_kwargs = {
    'name': 'kraken-wrapper',
    'version': '0.1.0',
    'description': '',
    'long_description': '# kraken-wrapper\n\nProvides the `krakenw` command which is a wrapper around Kraken to construct an isolated and reproducible build\nenvironment.\n\n__Features__\n\n* Produces isolated environments in PEX format\n* Reads build requirements from the `.kraken.py` file header\n* Produces lock files (`.kraken.lock`) that can be used to reconstruct an exact build environment <sup>1)</sup>\n\n<sup>1) The lock files do not contain hashes for installed distributions, but only the exact version numbers from\nthe resolved build environment.</sup>\n\n__Requirements header__\n\nIf no `.kraken.lock` file is present, Kraken wrapper will read the header of the `.kraken.py` file to obtain the\nrequirements to install into the build environment. The format of this header is demonstrated below:\n\n```py\n# ::requirements kraken-std>=0.3.0,<0.4.0 --extra-index-url https://...\n# ::pythonpath build-support\n```\n\nThe available options are:\n\n* **`requirements`**: Here you can specify any number of Pip requirements or local requirements (of the\n    format `dist-name @ path/to/dist`) as well as `--index-url`, `--extra-index-url` and `--interpreter-constraint`.\n* **`pythonpath`**: One or more paths to add the `sys.path` before your build script is executed. The `build-script` folder\n    is always added by default (as is the default behaviour by the `kraken-core` Python script project loader).\n\n',
    'author': 'Unknown',
    'author_email': 'me@unknown.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import argparse
import builtins
import logging
import sys
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, NoReturn

from . import __version__

if TYPE_CHECKING:
    from kraken.util.requirements import RequirementSpec

    from kraken.wrapper.buildenv import BuildEnvManager, BuildEnvType
    from kraken.wrapper.lockfile import Lockfile

BUILDENV_PATH = Path("build/.kraken/venv")
BUILDSCRIPT_FILENAME = ".kraken.py"
BUILD_SUPPORT_DIRECTORY = "build-support"
LOCK_FILENAME = ".kraken.lock"
_FormatterClass = lambda prog: argparse.RawTextHelpFormatter(prog, max_help_position=60, width=120)  # noqa: 731
logger = logging.getLogger(__name__)
print = partial(builtins.print, "[krakenw]", flush=True)
eprint = partial(print, file=sys.stderr)


def _get_argument_parser() -> argparse.ArgumentParser:
    from kraken._vendor.termcolor import colored
    from kraken.cli.option_sets import LoggingOptions
    from kraken.util.text import inline_text

    from kraken.wrapper.option_sets import EnvOptions

    parser = argparse.ArgumentParser(
        "krakenw",
        formatter_class=_FormatterClass,
        description=inline_text(
            f"""
            This is kraken-wrapper v{__version__}.

            {colored("krakenw", attrs=["bold"])} is a thin wrapper around the {colored("kraken", attrs=["bold"])} cli
            that executes builds in an isolated \\
            build environment. This ensures that builds are reproducible (especially when using \\
            lock files).

            To learn more about kraken, visit https://github.com/kraken-build/kraken-core.
            """
        ),
    )
    parser.add_argument("-V", "--version", version=__version__, action="version")
    LoggingOptions.add_to_parser(parser)
    EnvOptions.add_to_parser(parser)

    # NOTE (@NiklasRosenstein): If we combine "+" with remainder, we get options passed after the `cmd`
    #       passed directly into `args` without argparse treating it like an option. This is not the case
    #       when using `nargs=1` for `cmd`.
    parser.add_argument("cmd", nargs="*", metavar="cmd", help="{lock,l} or a kraken command")
    parser.add_argument("args", nargs=argparse.REMAINDER, help="additional arguments")
    return parser


def _get_lock_argument_parser(prog: str) -> argparse.ArgumentParser:
    from kraken._vendor.termcolor import colored
    from kraken.util.text import inline_text

    parser = argparse.ArgumentParser(
        prog,
        formatter_class=_FormatterClass,
        description=inline_text(
            f"""
            Rewrite the lock file ({colored(LOCK_FILENAME, attrs=["bold"])}) from the current build environment.
            """
        ),
    )

    return parser


def lock(prog: str, argv: list[str], manager: BuildEnvManager, requirements: RequirementSpec) -> NoReturn:
    from kraken.wrapper.lockfile import Lockfile

    parser = _get_lock_argument_parser(prog)
    parser.parse_args(argv)

    if not manager.exists():
        print("error: cannot lock without a build environment")
        sys.exit(1)

    environment = manager.get_environment()
    distributions = environment.get_installed_distributions()

    had_lockfile = Path(LOCK_FILENAME).exists()
    lockfile = Lockfile(requirements, {dist.name: dist.version for dist in distributions})
    lockfile.write_to(Path(LOCK_FILENAME))
    manager.set_locked(lockfile)

    eprint("lock file", "updated" if had_lockfile else "created", f"({LOCK_FILENAME})")
    sys.exit(0)


def _print_env_status(
    manager: BuildEnvManager,
    requirements: RequirementSpec,
    lockfile: Lockfile | None,
) -> None:
    from kraken.util.asciitable import AsciiTable
    from kraken.util.json import dt2json

    hash_algorithm = manager.get_hash_algorithm()

    table = AsciiTable()
    table.headers = ["Key", "Source", "Value"]
    table.rows.append(("Requirements", BUILDSCRIPT_FILENAME, requirements.to_hash(hash_algorithm)))
    if lockfile:
        table.rows.append(("Lockfile", LOCK_FILENAME, "-"))
        table.rows.append(("  Requirements hash", "", lockfile.requirements.to_hash(hash_algorithm)))
        table.rows.append(("  Pinned hash", "", lockfile.to_pinned_requirement_spec().to_hash(hash_algorithm)))
    else:
        table.rows.append(("Lockfile", LOCK_FILENAME, "n/a"))
    if manager.exists():
        metadata = manager.get_metadata()
        environment = manager.get_environment()
        table.rows.append(("Environment", str(environment.get_path()), environment.get_type().name))
        table.rows.append(("  Metadata", str(manager.get_metadata_file()), "-"))
        table.rows.append(("    Created at", "", dt2json(metadata.created_at)))
        table.rows.append(("    Requirements hash", "", metadata.requirements_hash))
    else:
        table.rows.append(("Environment", str(manager.get_environment().get_path()), "n/a"))
    table.print()


def _ensure_installed(
    manager: BuildEnvManager,
    requirements: RequirementSpec,
    lockfile: Lockfile | None,
    reinstall: bool,
    upgrade: bool,
    env_type: BuildEnvType | None = None,
) -> None:

    exists = manager.exists()
    install = reinstall or upgrade or not exists

    if not exists:
        env_type = env_type or env_type or manager.get_environment().get_type()
        eprint(f"initializing build environment (type: {env_type.name})")

    current_type = manager.get_environment().get_type()
    if env_type is not None and exists and env_type != current_type:
        eprint(f"re-initializing build environment (type changed: {current_type.name} → {env_type.name})")
        install = True

    if not install and exists:
        metadata = manager.get_metadata()
        if lockfile and metadata.requirements_hash != lockfile.to_pinned_requirement_spec().to_hash(
            metadata.hash_algorithm
        ):
            install = True
            eprint("re-initializing build environment (outdated compared to lockfile)")
        if not lockfile and metadata.requirements_hash != requirements.to_hash(metadata.hash_algorithm):
            install = True
            eprint("re-initializing build environment (outdated compared to requirements)")

    if install:
        if not lockfile or upgrade:
            logger.info("installing from requirements")
            source = requirements
        else:
            logger.info("installing from lock file")
            source = lockfile.to_pinned_requirement_spec()
            lockfile = None

        manager.remove()
        manager.install(source, env_type)


def main() -> NoReturn:
    from kraken.cli.option_sets import LoggingOptions
    from kraken.util.requirements import parse_requirements_from_python_script

    from kraken.wrapper.buildenv import BuildEnvManager
    from kraken.wrapper.lockfile import Lockfile
    from kraken.wrapper.option_sets import EnvOptions

    parser = _get_argument_parser()
    args = parser.parse_args()
    LoggingOptions.collect(args).init_logging()
    env_options = EnvOptions.collect(args)

    if not args.cmd and not env_options.any():
        parser.print_usage()
        sys.exit(0)

    if not Path(BUILDSCRIPT_FILENAME).is_file():
        print(f'error: no "{BUILDSCRIPT_FILENAME}" in current directory', file=sys.stderr)
        sys.exit(1)

    # Load requirement spec from build script.
    logger.debug('loading requirements from "%s"', BUILDSCRIPT_FILENAME)
    with Path(BUILDSCRIPT_FILENAME).open() as fp:
        requirements = parse_requirements_from_python_script(fp)
        if not requirements.requirements:
            print(f'error: no requirements in "{BUILDSCRIPT_FILENAME}"')
            sys.exit(1)
        if BUILD_SUPPORT_DIRECTORY not in requirements.pythonpath:
            requirements = requirements.with_pythonpath([BUILD_SUPPORT_DIRECTORY])

    # Load lockfile if it exists.
    if Path(LOCK_FILENAME).is_file():
        logger.debug('loading lockfile from "%s"', LOCK_FILENAME)
        lockfile = Lockfile.from_path(Path(LOCK_FILENAME))
        if lockfile.requirements != requirements:
            eprint(f'lock file "{LOCK_FILENAME}" is outdated compared to requirements in "{BUILDSCRIPT_FILENAME}"')
            eprint("consider updating the lock file with `krakenw --upgrade lock`")
    else:
        lockfile = None

    manager = BuildEnvManager(BUILDENV_PATH)

    if env_options.status:
        if args.cmd or args.args:
            eprint("error: --status option must be used alone")
            sys.exit(1)
        _print_env_status(manager, requirements, lockfile)
        sys.exit(0)

    if env_options.uninstall:
        if args.cmd or args.args:
            eprint("error: --uninstall option must be used alone")
            sys.exit(1)
        manager.remove()
        sys.exit(0)

    _ensure_installed(
        manager,
        requirements,
        lockfile,
        env_options.reinstall,
        env_options.upgrade,
        env_options.use,
    )

    cmd: str | None = args.cmd[0] if args.cmd else None
    argv: list[str] = args.cmd[1:] + args.args

    if cmd is None:
        assert not argv
        sys.exit(0)

    elif cmd in ("lock", "l"):
        lock(f"{parser.prog} lock", argv, manager, requirements)

    else:
        environment = manager.get_environment()
        environment.dispatch_to_kraken_cli([cmd, *argv])


if __name__ == "__main__":
    main()
